// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "exprs/vectorized/condition_expr.h"

#include "column/column_builder.h"
#include "column/column_helper.h"
#include "column/column_viewer.h"
#include "common/object_pool.h"
#include "exprs/vectorized/function_helper.h"

namespace starrocks {
namespace vectorized {

#define DEFINE_CLASS_CONSTRUCT_FN(NAME)         \
    NAME(const TExprNode& node) : Expr(node) {} \
                                                \
    ~NAME() {}                                  \
                                                \
    virtual Expr* clone(ObjectPool* pool) const override { return pool->add(new NAME(*this)); }

template <PrimitiveType Type>
class VectorizedIfNullExpr : public Expr {
public:
    DEFINE_CLASS_CONSTRUCT_FN(VectorizedIfNullExpr);

    ColumnPtr evaluate(ExprContext* context, vectorized::Chunk* ptr) override {
        auto lhs = _children[0]->evaluate(context, ptr);

        int null_count = ColumnHelper::count_nulls(lhs);
        if (null_count == 0) {
            return lhs->clone();
        }

        auto rhs = _children[1]->evaluate(context, ptr);
        if (null_count == lhs->size()) {
            return rhs->clone();
        }

        Columns list = {lhs, rhs};
        ColumnBuilder<Type> result(this->type().precision, this->type().scale);

        ColumnViewer<Type> lhs_viewer(lhs);
        ColumnViewer<Type> rhs_viewer(rhs);

        size_t size = list[0]->size();
        for (int row = 0; row < size; ++row) {
            if (lhs_viewer.is_null(row)) {
                result.append(rhs_viewer.value(row), rhs_viewer.is_null(row));
            } else {
                result.append(lhs_viewer.value(row), lhs_viewer.is_null(row));
            }
        }

        return result.build(ColumnHelper::is_all_const(list));
    }
};

template <PrimitiveType Type>
class VectorizedNullIfExpr : public Expr {
public:
    DEFINE_CLASS_CONSTRUCT_FN(VectorizedNullIfExpr);

    // NullIF: return null if lhs == rhs else return lhs
    ColumnPtr evaluate(ExprContext* context, vectorized::Chunk* ptr) override {
        auto lhs = _children[0]->evaluate(context, ptr);
        if (ColumnHelper::count_nulls(lhs) == lhs->size()) {
            return ColumnHelper::create_const_null_column(lhs->size());
        }

        auto rhs = _children[1]->evaluate(context, ptr);
        if (ColumnHelper::count_nulls(rhs) == rhs->size()) {
            return lhs->clone();
        }

        Columns list = {lhs, rhs};
        ColumnBuilder<Type> result(this->type().precision, this->type().scale);

        ColumnViewer<Type> lhs_viewer(lhs);
        ColumnViewer<Type> rhs_viewer(rhs);

        size_t size = list[0]->size();
        for (int row = 0; row < size; ++row) {
            if (lhs_viewer.is_null(row)) {
                result.append_null();
                continue;
            }

            if (!rhs_viewer.is_null(row) && (rhs_viewer.value(row) == lhs_viewer.value(row))) {
                result.append_null();
                continue;
            }

            result.append(lhs_viewer.value(row), lhs_viewer.is_null(row));
        }

        return result.build(ColumnHelper::is_all_const(list));
    }
};

template <PrimitiveType Type>
class VectorizedIfExpr : public Expr {
public:
    DEFINE_CLASS_CONSTRUCT_FN(VectorizedIfExpr);

    ColumnPtr evaluate(ExprContext* context, vectorized::Chunk* ptr) override {
        auto bhs = _children[0]->evaluate(context, ptr);
        int true_count = ColumnHelper::count_true_with_notnull(bhs);

        auto lhs = _children[1]->evaluate(context, ptr);
        if (true_count == bhs->size()) {
            return lhs->clone();
        }

        auto rhs = _children[2]->evaluate(context, ptr);
        if (true_count == 0) {
            return rhs->clone();
        }

        if (lhs->only_null() && rhs->only_null()) {
            return lhs->clone();
        }

        Columns list = {bhs, lhs, rhs};
        ColumnBuilder<Type> result(this->type().precision, this->type().scale);

        auto bhs_nulls = ColumnHelper::count_nulls(bhs);
        auto lhs_nulls = ColumnHelper::count_nulls(lhs);
        auto rhs_nulls = ColumnHelper::count_nulls(rhs);

        ColumnViewer<TYPE_BOOLEAN> bhs_viewer(bhs);
        ColumnViewer<Type> lhs_viewer(lhs);
        ColumnViewer<Type> rhs_viewer(rhs);
        size_t size = list[0]->size();

        // optimization for 3 columns all not null.
        if (bhs_nulls == 0 && lhs_nulls == 0 && rhs_nulls == 0) {
            for (int row = 0; row < size; ++row) {
                if (!bhs_viewer.value(row)) {
                    result.append(rhs_viewer.value(row));
                } else {
                    result.append(lhs_viewer.value(row));
                }
            }
        } else {
            for (int row = 0; row < size; ++row) {
                if (bhs_viewer.is_null(row) || !bhs_viewer.value(row)) {
                    if (rhs_viewer.is_null(row)) {
                        result.append_null();
                    } else {
                        result.append(rhs_viewer.value(row));
                    }
                } else {
                    if (lhs_viewer.is_null(row)) {
                        result.append_null();
                    } else {
                        result.append(lhs_viewer.value(row));
                    }
                }
            }
        }

        return result.build(ColumnHelper::is_all_const(list));
    }
};

template <PrimitiveType Type>
class VectorizedCoalesceExpr : public Expr {
public:
    DEFINE_CLASS_CONSTRUCT_FN(VectorizedCoalesceExpr);

    ColumnPtr evaluate(ExprContext* context, vectorized::Chunk* ptr) override {
        std::vector<ColumnViewer<Type>> viewers;
        std::vector<ColumnPtr> columns;
        for (int i = 0; i < _children.size(); ++i) {
            auto value = _children[i]->evaluate(context, ptr);
            auto null_count = ColumnHelper::count_nulls(value);

            // 1.return if first column is all not null.
            // 2.if there is a column all not null, It at least maybe the choice.
            // 3.don't need check if value is all null
            if (null_count == 0) {
                if (viewers.size() == 0) {
                    return value->clone();
                }

                // There is a column all not null.
                viewers.push_back(ColumnViewer<Type>(value));
                columns.push_back(value);

                // put it in viewers as last column.
                break;
            } else if (null_count != value->size()) {
                viewers.push_back(ColumnViewer<Type>(value));
                columns.push_back(value);
            }
        }

        // return only null
        if (columns.size() == 0) {
            return ColumnHelper::create_const_null_column(ptr != nullptr ? ptr->num_rows() : 1);
        }

        // direct return if only one
        if (columns.size() == 1) {
            // TODO: don't copy column if chunk support copy on write
            return columns[0]->clone();
        }

        // choose not null
        ColumnBuilder<Type> builder(this->type().precision, this->type().scale);
        int size = columns[0]->size();
        int col_size = viewers.size();

        for (int row = 0; row < size; ++row) {
            int col;
            for (col = 0; col < (col_size - 1); ++col) {
                // if not null
                if (!viewers[col].is_null(row)) {
                    builder.append(viewers[col].value(row));
                    break;
                }
            }

            // if last column
            if (col == (col_size - 1)) {
                if (viewers[col].is_null(row)) {
                    builder.append_null();
                } else {
                    builder.append(viewers[col].value(row));
                }
            }
        }

        return builder.build(ColumnHelper::is_all_const(columns));
    }
};

#undef DEFINE_CLASS_CONSTRUCT_FN

#define CASE_TYPE(TYPE, CLASS)        \
    case TYPE: {                      \
        return new CLASS<TYPE>(node); \
    }

#define CASE_ALL_TYPE(CLASS)           \
    CASE_TYPE(TYPE_BOOLEAN, CLASS);    \
    CASE_TYPE(TYPE_TINYINT, CLASS);    \
    CASE_TYPE(TYPE_SMALLINT, CLASS);   \
    CASE_TYPE(TYPE_INT, CLASS);        \
    CASE_TYPE(TYPE_BIGINT, CLASS);     \
    CASE_TYPE(TYPE_LARGEINT, CLASS);   \
    CASE_TYPE(TYPE_FLOAT, CLASS);      \
    CASE_TYPE(TYPE_DOUBLE, CLASS);     \
    CASE_TYPE(TYPE_CHAR, CLASS);       \
    CASE_TYPE(TYPE_VARCHAR, CLASS);    \
    CASE_TYPE(TYPE_DATE, CLASS);       \
    CASE_TYPE(TYPE_DATETIME, CLASS);   \
    CASE_TYPE(TYPE_DECIMALV2, CLASS);  \
    CASE_TYPE(TYPE_OBJECT, CLASS);     \
    CASE_TYPE(TYPE_HLL, CLASS);        \
    CASE_TYPE(TYPE_PERCENTILE, CLASS); \
    CASE_TYPE(TYPE_DECIMAL32, CLASS);  \
    CASE_TYPE(TYPE_DECIMAL64, CLASS);  \
    CASE_TYPE(TYPE_DECIMAL128, CLASS);

Expr* VectorizedConditionExprFactory::create_if_null_expr(const starrocks::TExprNode& node) {
    PrimitiveType resultType = TypeDescriptor::from_thrift(node.type).type;
    switch (resultType) {
        CASE_ALL_TYPE(VectorizedIfNullExpr);
    default:
        LOG(WARNING) << "vectorized engine not support type: " << resultType;
        return nullptr;
    }
}

Expr* VectorizedConditionExprFactory::create_null_if_expr(const TExprNode& node) {
    PrimitiveType resultType = TypeDescriptor::from_thrift(node.type).type;
    switch (resultType) {
        CASE_ALL_TYPE(VectorizedNullIfExpr);
    default:
        LOG(WARNING) << "vectorized engine not support type: " << resultType;
        return nullptr;
    }
}

Expr* VectorizedConditionExprFactory::create_if_expr(const TExprNode& node) {
    PrimitiveType resultType = TypeDescriptor::from_thrift(node.type).type;

    switch (resultType) {
        CASE_ALL_TYPE(VectorizedIfExpr);
    default:
        LOG(WARNING) << "vectorized engine not support type: " << resultType;
        return nullptr;
    }
}

Expr* VectorizedConditionExprFactory::create_coalesce_expr(const TExprNode& node) {
    PrimitiveType resultType = TypeDescriptor::from_thrift(node.type).type;
    switch (resultType) {
        CASE_ALL_TYPE(VectorizedCoalesceExpr);
    default:
        LOG(WARNING) << "vectorized engine not support type: " << resultType;
        return nullptr;
    }
}

#undef CASE_TYPE
#undef CASE_ALL_TYPE

} // namespace vectorized
} // namespace starrocks
